#!/usr/bin/python
# -*- coding: utf-8 -*-

import os
import subprocess
from shutil import rmtree
from multiprocessing import Pool

from postgres_ddl.Config import Config
from postgres_ddl.Database import Database
from postgres_ddl.System import SetupLogging

class Grabber():
    def __init__(self, config):
        self.Connect = config.get("connect") or {}

        self.Path = (config.get("path") or "").strip()
        assert len(self.Path) > 0, \
            "Grabber path is null"

        self.ExcludeSchemas = config.get("exclude_schemas") or []
        self.Threads = config.get("threads") or 8

        SetupLogging(config)

        Config.Parse(config)

    def NonSuperUserUndo(self):
        """
        Undo changes for superuser-only files
        """
        os.chdir(self.Path)
        res = subprocess.run(["git", "status", "--short"], shell=True, capture_output=True)
        res = res.stdout.decode("utf-8").replace(chr(10), chr(13))
        for file in res.split(chr(13)):
            file = file[3:]
            if file.find("_logical") >= 0:
                subprocess.run(["git", "checkout", "-q", "--", file], shell=True)

    def WriteFile(self, file):
        with open(file.get("path"), "w", encoding="utf8") as wf:
            wf.write(file.get("data"))

    def Run(self):
        database = Database(connect=self.Connect, exclude_schemas=self.ExcludeSchemas)
        database.Parse()

        if os.path.exists(self.Path):
            rmtree(self.Path)

        write_data = []
        for k,o in database.Objects.items():
            path = o.GetPath()
            if len(path) <= 0:
                continue

            file_name = o.GetFileName()
            if len(file_name) <= 0:
                continue

            path = [self.Path] + path
            path = "/".join(path)

            if not os.path.exists(path):
                os.makedirs(path)

            write_data.append({
                "path" : "/".join([path, file_name]),
                "data" : o.DDL_Create()
            })

        if (self.Threads or 1) <= 1:
            for file in write_data:
                self.WriteFile(file)
        else:
            with Pool(self.Threads) as pool:
                pool.map(self.WriteFile, write_data)

        if not database.IsSuperUser:
            self.NonSuperUserUndo()



#!/usr/bin/python
# -*- coding: utf-8 -*-

from postgres_ddl.Comment import Comment
from postgres_ddl.Config import Config
from postgres_ddl.DDL import DDL
from postgres_ddl.Grant import Grant
from postgres_ddl.Owner import Owner
from postgres_ddl.System import ParseACL

class ForeignServer(DDL):
    def __init__(self, parent, data):
        super().__init__(parent, data)

        self.Schema = "_foreign"

        self.Name = (data.get("server_name") or "").strip().lower()
        assert len(self.Name) > 0, \
            "Foreign server name is null"

        self.FDW = (data.get("fdw_name") or "").strip().lower()
        assert len(self.FDW) > 0, \
            "Foreign server FDW is null"

        self.Options = data.get("options")

        self.Owner = Owner(
            self.GetObjectName(),
            {
                "instance_type" : "SERVER",
                "instance_name" : self.Name,
                "owner_name"    : data.get("owner_name")
            }
        )

        self.Comment = Comment(
            self.GetObjectName(),
            {
                "instance_type" : self.GetTag(),
                "instance_name" : self.Name,
                "comment"       : data.get("comment")
            }
        )

        self.Grants = []
        for grant in ParseACL(data.get("acl"), self.Owner.Owner):
            grant["instance_type"] = self.GetTag()
            grant["instance_name"] = self.Name
            self.Grants.append(Grant(self.GetObjectName(), grant))

    def __str__(self):
        return self.Name

    def GetObjectType(self):
        return "foreign_server"

    def GetObjectName(self):
        return f"{self.GetObjectType()}_{self.Name}"

    def GetTag(self):
        return "FOREIGN SERVER"

    def DDL_Drop(self):
        return f"DROP SERVER IF EXISTS {self.Name};"

    def DDL_Create(self):
        r = f"-- Server: {self.Name}"
        r += Config.NL + Config.NL
        r += f"-- {self.DDL_Drop()}"
        r += Config.NL + Config.NL
        r += f"CREATE SERVER {self.Name}"
        r += Config.NL
        r += f"FOREIGN DATA WRAPPER {self.FDW}"
        r += Config.NL
        r += "OPTIONS("
        r += Config.NL
        r += self.DDL_Options()
        r += Config.NL
        r += ");"
        r += Config.NL + Config.NL
        r += self.Owner.DDL_Create()
        r += Config.NL + Config.NL

        for grant in self.Grants:
            r += grant.DDL_Create()
            r += Config.NL
        r += Config.NL

        if self.Comment.IsExists:
            r += self.Comment.DDL_Create()
            r += Config.NL

        return r.strip() + Config.NL

    def DDL_Options(self):
        result = []

        for o in sorted(self.Options):
            o = o.split("=")
            if len(o) != 2:
                continue
            result.append(f"{Config.Indent}{o[0]} = '{o[1]}'")

        return f",{Config.NL}".join(result)

    def GetPath(self):
        return [self.Schema]

    def GetFileName(self):
        return f"{self.Name}.sql"

    def Export(self):
        result = {}
        result[self.GetObjectName()] = self
        result[self.Comment.GetObjectName()] = self.Comment
        result[self.Owner.GetObjectName()] = self.Owner
        for v in self.Grants:
            result[v.GetObjectName()] = v
        return result



#!/usr/bin/python
# -*- coding: utf-8 -*-

from postgres_ddl.Config import Config
from postgres_ddl.DDL import DDL

class TableIndex(DDL):
    def __init__(self, parent, data):
        super().__init__(parent, data)

        self.Schema = (data.get("schema") or "").strip().lower()
        assert len(self.Schema) > 0, \
            "Index schema is null"

        self.Table = (data.get("table") or "").strip().lower()
        assert len(self.Table) > 0, \
            "Index table is null"

        self.Name = (data.get("name") or "").strip().lower()
        assert len(self.Name) > 0, \
            "Index name is null"

        self.Definition = (data.get("definition") or "")
        assert len(self.Definition) > 0, \
            "Index definition is null"

    def __str__(self):
        return self.GetFullName()

    def GetObjectType(self):
        return "table_index"

    def GetObjectName(self):
        return f"{self.GetObjectType()}_{self.GetFullName()}"

    def GetFullName(self):
        return f"{self.Schema}.{self.Name}"

    def GetTag(self):
        return "INDEX"

    def DDL_Create(self, separator=" "):
        definition = self.Definition
        definition = definition.replace(" ON ",    separator+"ON ")
        definition = definition.replace(" USING ", separator+"USING ")
        definition = definition.replace(" (",      separator+"(")
        definition = definition.replace(" WHERE",  separator+"WHERE")
        definition = definition + ";"

        r = f"-- Index: {self.GetFullName()}"
        r += Config.NL + Config.NL
        r += f"-- {self.DDL_Drop()}"
        r += Config.NL + Config.NL
        r += definition

        return r.strip()

    def DDL_Drop(self):
        return f"DROP {self.GetTag()} IF EXISTS {self.GetFullName()};"

    def Diff(self, another):
        if self.Definition != another.Definition:
            return [
                another.DDL_Drop(),
                self.DDL_Create()
            ]
        else:
            return []



#!/usr/bin/python
# -*- coding: utf-8 -*-

from postgres_ddl.Config import Config
from postgres_ddl.DDL import DDL

class Cast(DDL):
    def __init__(self, parent, data):
        super().__init__(parent, data)

        self.Oid = data.get("oid")
        assert (self.Oid or 0) > 0, \
            "Cast oid is null"

        self.TypeFrom = (data.get("type_from") or "").strip()
        assert len(self.TypeFrom) > 0, \
            "Cast type_from is null"

        self.TypeTo = (data.get("type_to") or "").strip()
        assert len(self.TypeTo) > 0, \
            "Cast type_to is null"

        self.Context = (data.get("context") or "").strip()
        assert len(self.Context) > 0, \
            "Cast context is null"

        self.Func = (data.get("func") or "").strip()

    def __str__(self):
        return self.GetFullName()

    def GetObjectType(self):
        return "cast"

    def GetObjectName(self):
        return f"{self.GetObjectType()}_{self.GetFullName()}"

    def GetFullName(self):
        return f"{self.TypeFrom} AS {self.TypeTo}"

    def GetTag(self):
        return "CAST"

    def DDL_Drop(self):
        return f"DROP {self.GetTag()} IF EXISTS {self.GetFullName()};"

    def DDL_Create(self):
        r = f"-- Cast: {self.GetFullName()}"
        r += Config.NL + Config.NL
        r += f"-- {self.DDL_Drop()}"
        r += Config.NL + Config.NL
        r += f"CREATE {self.GetTag()} ({self.GetFullName()})"
        r += Config.NL

        if len(self.Func) > 0:
            r += f"{Config.Indent}WITH FUNCTION {self.Func}"
        else:
            r += f"{Config.Indent}WITHOUT FUNCTION"
        r += Config.NL

        r += f"{Config.Indent}AS {self.Context}"
        r = r.strip() + ";"
        return r.strip() + Config.NL

    def GetPath(self):
        return ["_cast"]

    def GetFileName(self):
        return f"{self.GetFullName()}.sql"

    def Export(self):
        return { self.GetObjectName() : self }

    def Diff(self, another):
        if self.DDL_Create() != another.DDL_Create():
            return [
                another.DDL_Drop(),
                self.DDL_Create()
            ]

        return []



#!/usr/bin/python
# -*- coding: utf-8 -*-

class Config():
    Indent = " " * 2
    NL = chr(10)

    @staticmethod
    def Parse(json):
        Config.Indent = " " * (json.get("indent") or 2)
        Config.NL = json.get("new_line") or chr(10)



#!/usr/bin/python
# -*- coding: utf-8 -*-

from postgres_ddl.Config import Config
from postgres_ddl.DDL import DDL
from postgres_ddl.Owner import Owner

class Subscription(DDL):
    def __init__(self, parent, data):
        super().__init__(parent, data)

        self.Oid = data.get("oid")
        assert (self.Oid or 0) > 0, \
            "Subscription oid is null"

        self.Name = (data.get("name") or "").strip()
        assert len(self.Name) > 0, \
            "Subscription name is null"

        self.IsEnabled = bool(data.get("is_enabled") or False)

        self.Connect = (data.get("connect") or "").strip()
        assert len(self.Connect) > 0, \
            "Subscription connect is null"

        self.Slot = (data.get("slot") or "").strip()
        assert len(self.Slot) > 0, \
            "Subscription slot is null"

        self.SyncCommit = (data.get("sync_commit") or "").strip()
        assert len(self.SyncCommit) > 0, \
            "Subscription sync_commit is null"

        self.Publications = (data.get("publications") or [])
        assert len(self.Publications) > 0, \
            "Subscription publications is null"

        self.Tables = (data.get("tables") or [])

        self.Owner = Owner(
            self.GetObjectName(),
            {
                "instance_type" : self.GetTag(),
                "instance_name" : self.GetFullName(),
                "owner_name"    : data.get("owner")
            }
        )

    def __str__(self):
        return self.GetFullName()

    def GetObjectType(self):
        return "subscription"

    def GetObjectName(self):
        return f"{self.GetObjectType()}_{self.GetFullName()}"

    def GetFullName(self):
        return self.Name

    def GetTag(self):
        return "SUBSCRIPTION"

    def DDL_Drop(self):
        return f"DROP {self.GetTag()} IF EXISTS {self.GetFullName()};"

    def DDL_Create(self):
        r = f"-- Subscription: {self.GetFullName()}"
        r += Config.NL + Config.NL
        r += f"-- {self.DDL_Drop()}"
        r += Config.NL + Config.NL
        r += f"CREATE {self.GetTag()} {self.GetFullName()}"
        r += Config.NL
        r += f"CONNECTION '{self.Connect}'"
        r += Config.NL
        r += f"PUBLICATION {self.GetPublications()}"
        r += Config.NL
        r += "WITH ("
        r += Config.NL
        r += f"{Config.Indent}enabled = {str(self.IsEnabled).upper()},"
        r += Config.NL
        r += f'{Config.Indent}slot_name = "{self.Slot}",'
        r += Config.NL
        r += f"{Config.Indent}synchronous_commit = '{self.SyncCommit}'"
        r += Config.NL
        r += ");"
        r += Config.NL + Config.NL
        r += self.Owner.DDL_Create() + Config.NL

        if len(self.Tables) > 0:
            r += Config.NL
            r += '-- Tables:' + Config.NL
            for t in self.Tables:
                r += f'--   {t}'
                r += Config.NL

        return r.strip() + Config.NL

    def GetPath(self):
        return ["_logical"]

    def GetFileName(self):
        return f"{self.Name}.sql"

    def GetPublications(self):
        return ", ".join(self.Publications)

    def Export(self):
        result = {}
        result[self.GetObjectName()] = self
        result[self.Owner.GetObjectName()] = self.Owner
        return result

    def Diff(self, another):
        result = []

        if self.IsEnabled != another.IsEnabled:
            if self.IsEnabled:
                result.append(f"ALTER {self.GetTag()} {self.GetFullName()} ENABLE;")
            else:
                result.append(f"ALTER {self.GetTag()} {self.GetFullName()} DISABLE;")

        if self.Connect != another.Connect:
            result.append(f"ALTER {self.GetTag()} {self.GetFullName()} CONNECTION '{self.Connect}';")

        if self.Publications != another.Publications:
            result.append(f"ALTER {self.GetTag()} {self.GetFullName()} SET PUBLICATION {self.GetPublications()};")

        if self.Slot != another.Slot:
            result.append(f'ALTER {self.GetTag()} {self.GetFullName()} SET(slot_name = "{self.Slot}");')

        if self.Slot != another.Slot:
            result.append(f"ALTER {self.GetTag()} {self.GetFullName()} SET(synchronous_commit = '{self.SyncCommit}');")

        if len(result) > 0:
            result.append(f"ALTER {self.GetTag()} {self.GetFullName()} REFRESH PUBLICATION;")

        return result




#!/usr/bin/python
# -*- coding: utf-8 -*-

from postgres_ddl.DDL import DDL

class Grant(DDL):
    def __init__(self, parent, data):
        super().__init__(parent, data)

        self.Type = (data.get("instance_type") or "").strip().upper()
        assert len(self.Type) > 0, \
            f"Grant instance type is null - {parent}"

        self.Instance = (data.get("instance_name") or "").strip()
        assert len(self.Instance) > 0, \
            f"Grant instance name is null - {parent}"

        self.IsGrant = data.get("is_grant")
        assert self.IsGrant is not None, \
            f"Grant type is null - {parent}"

        assert isinstance(self.IsGrant, bool), \
            f"Grant type is not boolean - {parent}"

        self.Status = "GRANT" if self.IsGrant else "REVOKE"
        self.StatusDrop = "GRANT" if not self.IsGrant else "REVOKE"

        self.Prep = "TO" if self.IsGrant else "FROM"
        self.PrepDrop = "TO" if not self.IsGrant else "FROM"

        self.Permissions = data.get("perm")
        assert self.Permissions is not None, \
            f"Grant permissions is null - {parent}"

        assert isinstance(self.Permissions, list), \
            f"Grant permissions is not list - {parent}"

        assert len(self.Permissions) > 0, \
            f"Grant permissions list is empty - {parent}"

        self.Permissions = ", ".join(sorted(self.Permissions))

        self.Role = (data.get("role") or "").strip()
        assert len(self.Role) > 0, \
            "Grant role is null"

    def __str__(self):
        return self.GetObjectName()

    def GetObjectType(self):
        return "grant"

    def GetObjectName(self):
        return "_".join([self.GetObjectType(), self.Type, self.Instance, self.Role])

    def DDL_Create(self):
        return "%s %s ON %s %s %s %s;" % (
            self.Status, self.Permissions, self.Type,
            self.Instance, self.Prep, self.Role
        )

    def DDL_Drop(self):
        return "%s %s ON %s %s %s %s;" % (
            self.StatusDrop, self.Permissions, self.Type,
            self.Instance, self.PrepDrop, self.Role
        )

    def Diff(self, another):
        if self.Permissions != another.Permissions:
            return [
                another.DDL_Drop(),
                self.DDL_Create()
            ]
        return []
